package de.unibi.airobots.bonsai.sensors;

import java.io.IOException;

import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.btl.data.FaceIdentificationList;

/**
 * An interface for sensors returning face identification results. Moreover
 * learning new persons can be triggered with this interface.
 * 
 * Returning the identification results may be costly so that it can be enabled
 * via {@link #activate()} or disabled with {@link #deactivate()}. Learning is
 * only enabled if the system was activated before via {@link #activate()}.
 * 
 * @author jwienke
 */
public interface FaceIdentificationSensor extends
        Sensor<FaceIdentificationList> {

    /**
     * Activates the identification mechanism. This should only be done if
     * necessary because the computational costs are high.
     * 
     * @throws IOException
     *             communication error
     */
    void activate() throws IOException;

    /**
     * Activates the identification mechanism. This should only be done if
     * necessary because the computational costs are high.
     * 
     * @throws IOException
     *             communication error
     */
    void deactivate() throws IOException;

    /**
     * Enables the cropping of the left of the two faces with highest
     * reliability.
     * 
     * @throws IOException
     *             communication error
     */
    void twoFaces() throws IOException;
    
    
    /**
     * Starts learning the person in front of the camera with a user supplied
     * id. Learning can be stopped with {@link #stopLearning()}. While learning
     * {@link #readLast()} will (after some time) return the supplied id.
     * 
     * @param id
     *            the id of the person to learn
     * @throws IOException
     *             communication error
     */
    void startLearning(int id) throws IOException;

    /**
     * Stops learning the last supplied id.
     * 
     * @throws IOException
     *             communication error
     */
    void stopLearning() throws IOException;
    
    /**
     * Reset.
     * 
     * @throws IOException
     *             communication error
     */
    void reset() throws IOException;

}
