package de.unibi.airobots.bonsai.util;

import de.unibi.airobots.btl.data.NavigationGoalData;
import de.unibi.airobots.btl.data.PositionData;

/**
 * Utility class with methods to convert between several coordinate systems.
 */
public class CoordinateSystemConverter {

    /**
     * Generate global carthesian coordinates suitable as a navigation goal for
     * the NavigationActuator from local polar coordinate system coordinates
     * given as distance and angle. This method requires the current global
     * position (OdometryData) of the robot obtained from the OdometrySensor.
     * 
     * 
     * @param positionData
     *            global position of the robot as recently obtained from an
     *            OdometrySensor
     * @param angle
     *            angle of the polar coordinates in radiants (between 0 and 2 *
     *            Math.PI). An angle=0 corresponds to a position directly in
     *            front of the robot. angles are defined counter-clockwise.
     * @param dist
     *            distance of polar coordinates in metres (always >0).
     * @return the global carthesian position in metres and the corresponding
     *         orientation of the robot as NavigationGoalData, which can
     *         directly by used for the NavigationActuator.
     */
    public static NavigationGoalData polar2NavigationGoalData(
            PositionData positionData, double angle, double dist) {
        NavigationGoalData cart = new NavigationGoalData();
        double localAngle = angle + positionData.getYaw();

        if (localAngle > 2 * Math.PI) {
            localAngle = localAngle - 2 * Math.PI;
        }
        if (localAngle < 0) {
            localAngle = localAngle + 2 * Math.PI;
        }
        cart.setX(positionData.getX() + Math.cos(localAngle) * dist);

        cart.setY(positionData.getY() + Math.sin(localAngle) * dist);
        cart.setYaw(localAngle);
        return cart;
    }
    
    public static double positionData2Angle(
            PositionData ownPosition, PositionData objectPosition) {
        
        double angle = Math.atan2(objectPosition.getY() - ownPosition.getY(), objectPosition.getX() - ownPosition.getX());
        
        angle += ownPosition.getYaw();
        
        if (angle >= Math.PI) {
            angle -= 2*Math.PI;
        }
        if (angle < -Math.PI) {
            angle += 2* Math.PI;
        }
        
        return angle;
    }
    
    public static double positionDistance(PositionData ownPosition, PositionData objectPosition) {
        
        double x = ownPosition.getX() - objectPosition.getX();
            
        double y = ownPosition.getY() - objectPosition.getY();
        
        double dist = Math.sqrt( Math.pow(x, 2) + Math.pow(y, 2) );
        
        return dist;
    }

}
