package de.unibi.airobots.bonsai.util.helper;

import java.io.IOException;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.sensors.data.PersonList;
import de.unibi.airobots.btl.data.PersonData;

public class PersonHelpers {

	private Logger logger = Logger.getLogger(PersonHelpers.class);

	public static double FRONT_MAX_DISTANCE = 1.2;
	public static double FRONT_MAX_ANGLE = Math.PI / 4.0;

	private Sensor<PersonList> personSensor;

	private HashMap<String, String> personPositions = null;

	public PersonHelpers(Sensor<PersonList> personSensor) {
		personPositions = new HashMap<String, String>();
		this.personSensor = personSensor;
	}

	/**
	 * Returns the next person that somehow stands in front of the robot
	 * 
	 * @return person
	 * @throws IllegalStateException
	 *             no person in front
	 * @throws InterruptedException
	 */
	public PersonData getNextPersonInFront() throws IllegalStateException,
			InterruptedException {

		try {

			PersonList persons = personSensor.readLast();

			if (persons.size() == 0) {
				throw new IllegalStateException("No persons found.");
			}

			List<PersonData> personList = new LinkedList<PersonData>();
			for (PersonData p : persons) {
				if (isStandingInFront(p)) {
					personList.add(p);
				}
			}
			sortPersonsByDistance(personList);

			if (personList.size() == 0) {
				throw new IllegalStateException(
						"No person in front of robot found.");
			}

			return personList.get(0);

		} catch (IOException e) {
			logger.warn("IOError", e);
			throw new IllegalStateException("No persons found (IOException)", e);
		}

	}

	private void sortPersonsByDistance(List<PersonData> persons) {
		Collections.sort(persons, new Comparator<PersonData>() {

			@Override
			public int compare(PersonData o1, PersonData o2) {
				return new Double(o1.getDistance()).compareTo(o2.getDistance());
			}

		});
	}

	public void addPersonAndPosition(String personName, String personPosition) {
		this.personPositions.put(personName, personPosition);
	}

	private boolean isStandingInFront(PersonData p) {

		// TODO use facing information

		if (p.getDistance() > FRONT_MAX_DISTANCE) {
			logger.debug("Person too far away");
			return false;
		}

		if (Math.cos(p.getAngle()) < Math.cos(FRONT_MAX_ANGLE)) {
			logger.debug("Person not in right angle");
			return false;
		}

		if (p.isWalking()) {
			logger.debug("Person is walking");
			return false;
		}

		return true;

	}

	public HashMap<String, String> getPersonPositions() {
		return personPositions;
	}

}
