/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package de.unibi.airobots.bonsai.util.xml;

import java.awt.Polygon;
import java.awt.Rectangle;
import java.io.FileOutputStream;
import java.io.IOException;

import javax.vecmath.Vector3d;

import nu.xom.Attribute;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Serializer;
import de.unibi.airobots.bonsai.util.helper.annotation.MapArea;
import de.unibi.airobots.bonsai.util.helper.annotation.MapPoint;
import de.unibi.airobots.bonsai.util.helper.annotation.MapRegion;
import de.unibi.airobots.btl.data.NavigationGoalData;

/**
 * Some common encoding tasks.
 * 
 * @author Ingo Luetkebohle <iluetkeb@techfak.uni-bielefeld.de>
 */
public class XMLEncoderUtil {
    private static final String UTF_8 = "utf-8";
    /** Attribute name for values on the x-axis. */
    public static final String X_ATTRNAME = "x";
    /** Attribute name for values on the y-axis. */
    public static final String Y_ATTRNAME = "y";
    /** Attribute name for values on the z-axis. */
    public static final String Z_ATTRNAME = "z";
    /** Attribute name for widths of rectangles. */
    public static final String W_ATTRNAME = "width";
    /** Attribute name for heights of rectangles. */
    public static final String H_ATTRNAME = "height";
    /** Attribute name for points of polygons. */
    public static final String POINT_ATTRNAME = "POINT";
    
    public static final String TYPE_ATTRNAME = "type";
    
    public static final String NAME_ATTRNAME = "name";
    public static final String LABEL_ATTRNAME = "label";
    public static final String GROUP_ATTRNAME = "group";
    public static final String HIGH_ATTRNAME = "high";
    
    public static final String KIND_ATTRNAME = "kind";
    public static final String UNIT_ATTRNAME = "unit";
    public static final String REF_ATTRNAME = "ref";
    
    public static final String REGION_ELEMENTNAME = "REGION";
    public static final String COORD_ELEMENTNAME = "COORD";
    public static final String VIEW_POINT_ELEMENTNAME = "VIEW_POINT";
    public static final String YAW_ATTRNAME = "yaw";

    /**
     * Create an attribute with the given name and the string representation of
     * the given integer as its value.
     */
    protected static final void writeInt(Element e, String attrName, int v) {
        e.addAttribute(new Attribute(attrName, Integer.toString(v)));
    }

    /**
     * Create an attribute with the given name and the string representation of
     * the given integer as its value.
     */
    protected static final void writeDouble(Element e, String attrName, double v) {
        e.addAttribute(new Attribute(attrName, Double.toString(v)));
    }

    public static final void writePolygon(Element e, Polygon p) {
        int[] xPoints = p.xpoints;
        int[] yPoints = p.ypoints;
        
        int xlength = xPoints.length;
        if (xlength != yPoints.length) {
            return;
        }
  
        for (int i = 0; i < xlength; i++) {
            Element point = new Element(POINT_ATTRNAME);
            writeInt(point, X_ATTRNAME, xPoints[i]);
            writeInt(point, Y_ATTRNAME, yPoints[i]);
            e.appendChild(point);
        }
        
    }
    
    /** Writer integer rectangle into the attributes of the given element. */
    public static final void writeRectangle(Element e, Rectangle r) {
        writeInt(e, X_ATTRNAME, r.x);
        writeInt(e, Y_ATTRNAME, r.y);
        writeInt(e, W_ATTRNAME, r.width);
        writeInt(e, H_ATTRNAME, r.height);
    }

    /** Writer world coordinate into the attributes of the given element. */
    public static final void writeWorldCoord(Element e, Vector3d v) {
        writeDouble(e, X_ATTRNAME, v.x);
        writeDouble(e, Y_ATTRNAME, v.y);
        writeDouble(e, Z_ATTRNAME, v.z);
    }
    
    public static final void writeMapArea(Element e, MapArea area){
        
        Attribute type = new Attribute(TYPE_ATTRNAME, area.getType().name());
        Attribute name = new Attribute(NAME_ATTRNAME, area.getArenaName() );
        
        e.addAttribute(type);
        e.addAttribute(name);
        
        for (MapRegion region: area.getRegions()) {
            Element regionElement = new Element(REGION_ELEMENTNAME);
            writeMapRegion(regionElement, region);
            e.appendChild(regionElement);
        }
        
    }
    
    private static void writeMapRegion(Element e, MapRegion region) {
        Attribute type = new Attribute(TYPE_ATTRNAME, region.getType());
        Attribute label = new Attribute(LABEL_ATTRNAME, region.getLabel() );
        Attribute group = new Attribute(GROUP_ATTRNAME, region.getGroup());
        Attribute high = new Attribute(HIGH_ATTRNAME, String.valueOf(region.getHeight()));
        
        e.addAttribute(type);
        e.addAttribute(label);
        e.addAttribute(group);
        e.addAttribute(high);
        
        if (region.getViewPoint() != null){
        
            Element viewPointElement = new Element(VIEW_POINT_ELEMENTNAME);
            writePointElement(viewPointElement, region.getViewPoint());
            e.appendChild(viewPointElement);
        }
        
        for (MapPoint coord: region.getPoints()){
            Element coordElement = new Element(COORD_ELEMENTNAME);
            writeMapPoint(coordElement, coord );
            e.appendChild(coordElement);
        }
        
    }

    private static void writePointElement(Element e,
            NavigationGoalData viewPoint) {
        
        writeDouble(e, X_ATTRNAME, viewPoint.getX());
        writeDouble(e, Y_ATTRNAME, viewPoint.getY());
        writeDouble(e,  YAW_ATTRNAME,  viewPoint.getYaw());
        
    }
    
    private static void writeMapPoint(Element e, MapPoint coord) {
        writeDouble(e, X_ATTRNAME, coord.getX());
        writeDouble(e, Y_ATTRNAME, coord.getY());
        e.addAttribute(new Attribute( KIND_ATTRNAME, coord.getKind().name()));
        e.addAttribute(new Attribute( UNIT_ATTRNAME, coord.getUnit().name()));
        e.addAttribute(new Attribute( REF_ATTRNAME, coord.getRef()));
        
    }

    public static final void xmlWriter(Element root, String path) throws IOException {
        
        Document doc = new Document(root);
        
        FileOutputStream outPutStream = new FileOutputStream(path);
        
        Serializer serializer = new Serializer(outPutStream, UTF_8);
        serializer.setIndent(4);
        serializer.setMaxLength(64);
        serializer.write(doc);  

        
    }
}
