package de.unibi.airobots.bonsai.xcf;

import java.util.Map;

import Ice.ConnectFailedException;
import net.sf.xcf.RemoteServer;

/**
 * Interface for classes that work on {@link RemoteServer}s with the ability to
 * switch the remote server instance while working. This may be the case if that
 * server dies and is recreated afterwards.
 * 
 * <p>
 * Implementing classes may decide between various error handling strategies for
 * the remote servers they will receive. These strategies are defined in
 * {@link ErrorHandlingPolicy}.
 * </p>
 * 
 * @author jwienke
 */
public interface RemoteServerReconnectable {

    /**
     * Possible error handling policies for remote server connection errors.
     * 
     * @author jwienke
     */
    enum ErrorHandlingPolicy {

        /**
         * A requested remote server for this error handling strategy tries
         * everything it can do to handle connection errors on its own. This eg.
         * means that method calls are postponed until a new connection is
         * established and automatically re-executed then until they succeed.
         * The client of such a remote server will not know anything about
         * possible connection losses. This policy can be used in a
         * "select and forget" manner but may not be sufficient in every case.
         * This may be true if eg. a state must be restored after a connection
         * loss.
         */
        SELF_HEALING,

        /**
         * A requested remote server with this error handling policy will pass
         * all exception that occurred due to connection errors to the client.
         * After a connection error (usually a subclass of
         * {@link ConnectFailedException}) this remote server will generally not
         * be usable until it is reset via
         * {@link RemoteServerReconnectable#resetRemoteServer(RemoteServer)}.
         * Use this policy when the client needs to be aware of connection
         * errors.
         */
        NOTIFYING

    }

    /**
     * Returns the name of the remote server this class needs to work on and the
     * error handling policies requested per remote server.
     * 
     * @see ErrorHandlingPolicy
     * @return remote server names associated with the desired
     *         {@link ErrorHandlingPolicy}
     */
    Map<String, ErrorHandlingPolicy> getRemoteServerNames();

    /**
     * Resets the remote server or initializes it.
     * 
     * Implementing classes can use this method to restore the state of the
     * remote server based on the state of the old remote server. E.g. this may
     * be useful if a server has a method that enables another subscriber also
     * registered on. If the server is recreated and the subscriber is most
     * likely disabled and must be started again.
     * 
     * @param remoteServer
     *            new remote server instance to use
     */
    void resetRemoteServer(RemoteServer remoteServer);

}
