package de.unibi.airobots.bonsai.xcf.actuators;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.xcf.MethodNotFoundException;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.UserException;
import net.sf.xcf.transport.XOPData;
import nu.xom.Document;
import nu.xom.Element;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.actuators.SpeechActuator;
import de.unibi.airobots.bonsai.xcf.RemoteServerReconnectable;

/**
 * Actuator which can be used to access the speech synthesizer.
 * 
 * @author lschilli
 * @author jwienke
 * @author arothert
 */
public class XcfSpeechActuator implements SpeechActuator,
        RemoteServerReconnectable {

    private Logger logger = Logger.getLogger(getClass());

    private String remoteServerName = "";
    private Object remoteServerLock = new Object();
    private RemoteServer remoteServer = null;

    /**
     * Constructor.
     * 
     * @param remoteServerName
     *            name of the remote server used to produce speech
     */
    public XcfSpeechActuator(String remoteServerName) {
        this.remoteServerName = remoteServerName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void say(String text, boolean async) throws IOException {

        if(isNotAllowedSayChar(text)) {
            
            synchronized (remoteServerLock) {
                if (remoteServer == null) {
                    throw new IllegalStateException("No remote server set.");
                }
            }
    
            Element elem = new Element("raw");
            elem.appendChild(text);
            Document doc = new Document(elem);
            try {
                if (async) {
                    synchronized (remoteServerLock) {
                        remoteServer.callMethod("async_say_raw", new XOPData(doc));
                    }
                } else {
                    synchronized (remoteServerLock) {
                        remoteServer.callMethod("say_raw", new XOPData(doc));
                    }
                }
            } catch (MethodNotFoundException e) {
                logger.error(e);
                throw new IOException(e);
            } catch (UserException e) {
                logger.error(e);
                throw new IOException(e);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void say(String text) throws IOException {
        say(text, false);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Map<String, ErrorHandlingPolicy> getRemoteServerNames() {
        Map<String, ErrorHandlingPolicy> servers =
                new HashMap<String, ErrorHandlingPolicy>();
        servers.put(remoteServerName, ErrorHandlingPolicy.SELF_HEALING);
        return servers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetRemoteServer(RemoteServer remoteServer) {
        assert remoteServer.getServerName().equals(remoteServerName);
        synchronized (remoteServerLock) {
            this.remoteServer = remoteServer;
        }
    }

    @Override
    public void sayAccentuated(String accentuatedText, boolean async) throws IOException {
        
        if(isNotAllowedSayChar(accentuatedText)) {
        
            synchronized (remoteServerLock) {
                if (remoteServer == null) {
                    throw new IllegalStateException("No remote server set.");
                }
            }
            
            Element elem = new Element("accentuatedText");
            Element text = new Element("text");
            Element bool = new Element("bool");
            
            text.appendChild(accentuatedText);
            bool.appendChild(Boolean.toString(async));
            elem.appendChild(text);
            elem.appendChild(bool);
            
            Document doc = new Document(elem);
            try {
                synchronized (remoteServerLock) {
                    remoteServer.callMethod("say_accented", new XOPData(doc));
                }
            } catch (MethodNotFoundException e) {
                logger.error(e);
                throw new IOException(e);
            } catch (UserException e) {
                logger.error(e);
                throw new IOException(e);
            }
        }
    }

    @Override
    public void sayAccentuated(String accentuatedText, boolean async,
            String prosodyConfig) throws IOException {
        
        if(isNotAllowedSayChar(accentuatedText)) {
            
            synchronized (remoteServerLock) {
                if (remoteServer == null) {
                    throw new IllegalStateException("No remote server set.");
                }
            }
            
            Element elem = new Element("accentuatedText");
            Element text = new Element("text");
            Element bool = new Element("bool");
            Element prosody = new Element("prosody");
            
            text.appendChild(accentuatedText);
            bool.appendChild(Boolean.toString(async));
            prosody.appendChild(prosodyConfig);
            
            elem.appendChild(text);
            elem.appendChild(bool);
            elem.appendChild(prosody);
            
            Document doc = new Document(elem);
            try {
                synchronized (remoteServerLock) {
                    remoteServer.callMethod("say_accented", new XOPData(doc));
                }
            } catch (MethodNotFoundException e) {
                logger.error(e);
                throw new IOException(e);
            } catch (UserException e) {
                logger.error(e);
                throw new IOException(e);
            }
        }
    }
    
    public void sayAccentuated(String accented_text) throws IOException{
        sayAccentuated(accented_text, false);
    }
    
    public void sayAccentuated(String accented_text, String prosodyConfig) throws IOException{
        sayAccentuated(accented_text, false, prosodyConfig);
    }

    /**
     * check if string entails signs which through a null-pointer exception
     * @param str say string which should be spoken
     * @return true if the string is allowed else false
     */
    private boolean isNotAllowedSayChar(String str){
        
        Pattern p = Pattern.compile("[ ]*?[.,!?;:()]+");
        Matcher m = p.matcher(str);

        if (m.matches() || str.equals("")) {
            logger.error("String to say is one of the following symbols: ,.!?():; see: " + str);
                return false;
        }
        return true;
    }
    
}
