package de.unibi.airobots.bonsai.xcf.configuration;

import java.util.Map;

import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.core.CoreObjectFactory.SensorToConfigure;

/**
 * Base class for XCF-based sensor configuration objects. The default
 * implementation of {@link #parseOptions(Map)} requires a publisher name for
 * convenience reasons.
 * 
 * @author jwienke
 * @param <T>
 *            data type returned by the sensor
 */
public abstract class ConfiguredXcfSensor<T> extends ConfiguredXcfCoreObject {

    /**
     * Option for publisher based sensors defining the publisher name.
     */
    public static final String OPTION_PUBLISHER_NAME = "publisherName";

    /**
     * Data type to be returned by the sensor.
     */
    private Class<T> dataType = null;

    /**
     * Name of the configured publisher if {@link #parseOptions(Map)} was used.
     */
    private String publisherName = "";

    /**
     * Create a new specific sensor instance.
     * 
     * @return new instance
     */
    public abstract Sensor<T> createInstance();

    /**
     * {@inheritDoc}
     * 
     * Default implementation parsing the publisher name.
     */
    @Override
    public void parseOptions(Map<String, String> options)
            throws IllegalArgumentException {

        if (!options.containsKey(OPTION_PUBLISHER_NAME)) {
            throw new IllegalArgumentException(
                    "Publisher name missing in arguments.");
        }

        publisherName = options.get(OPTION_PUBLISHER_NAME);

    }

    /**
     * Override this and check for suitable data type!
     * 
     * @param sensor
     *            requested sensor
     * @return <code>true</code> if configuration can create such a requested
     *         sensor
     */
    public static boolean isSuitableFor(SensorToConfigure sensor) {
        return false;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String[] getSubscribersToInitialize() {
        return new String[] { publisherName };
    }

    /**
     * Sets the data type the configured sensor is intended to return.
     * 
     * @param dataType
     *            new data type
     */
    public void setDataType(Class<T> dataType) {
        this.dataType = dataType;
    }

    /**
     * Returns the data type the configured sensor will return.
     * 
     * @return data type class
     */
    public Class<T> getDataType() {
        return dataType;
    }

    /**
     * Returns the configured publisher name for the sensor if
     * {@link #parseOptions(Map)} was used.
     * 
     * @return publisher name
     */
    public String getPublisherName() {
        return publisherName;
    }

}
