package de.unibi.airobots.bonsai.xcf.sensors;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import net.sf.xcf.MethodNotFoundException;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.UserException;
import net.sf.xcf.transport.XOPData;
import net.sf.xcf.xml.SyntaxException;
import de.unibi.airobots.bonsai.sensors.FaceIdentificationSensor;
import de.unibi.airobots.bonsai.xcf.RemoteServerReconnectable;
import de.unibi.airobots.btl.data.FaceIdentificationList;

/**
 * A sensor for face identification results provided by caiwicat.
 * 
 * @author jwienke
 */
public class XcfFaceIdentificationSensor extends
        XcfBtlPublisherSensor<FaceIdentificationList> implements
        RemoteServerReconnectable, FaceIdentificationSensor {

    // TODO these method could also be made configurable in the configuration
    // object

    // activation server names
    private static final String ACTIVATE_METHOD_NAME = "activate";
    private static final String DEACTIVATE_METHOD_NAME = "deactivate";

    // control server names
    private static final String START_LEARNING_METHOD_NAME = "startLearning";
    private static final String STOP_LEARNING_METHOD_NAME = "stopLearning";
    private static final String RESET_METHOD_NAME = "reset";
    private static final String TWOFACES_METHOD_NAME = "twofaces";

    private String activationServerName = "";
    private Object activationServerLock = new Object();
    /**
     * Server used to activate and deactivate the identification mechanism.
     */
    private RemoteServer activationServer = null;

    private String controlServerName = "";
    private Object controlServerLock = new Object();
    /**
     * Server used to control the learning process.
     */
    private RemoteServer controlServer;

    /**
     * Constructor.
     * 
     * @param resultSubscriberName
     *            name of the subscriber that receives the face identification
     *            results
     * @param activationServerName
     *            name of the server to activate or deactivate identification
     * @param controlServerName
     *            name of the server used to control the identification and
     *            learning process
     */
    public XcfFaceIdentificationSensor(String resultSubscriberName,
            String activationServerName, String controlServerName) {

        super(FaceIdentificationList.class, resultSubscriberName);

        this.activationServerName = activationServerName;
        this.controlServerName = controlServerName;

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void activate() throws IOException {

        try {
            synchronized (activationServerLock) {
                activationServer.callMethod(ACTIVATE_METHOD_NAME, XOPData
                        .fromString("<activate/>"));
            }
        } catch (MethodNotFoundException e) {
            throw new IOException(e);
        } catch (UserException e) {
            throw new IOException(e);
        } catch (SyntaxException e) {
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void deactivate() throws IOException {

        try {
            synchronized (activationServerLock) {
                activationServer.callMethod(DEACTIVATE_METHOD_NAME, XOPData
                        .fromString("<deactivate/>"));
            }
        } catch (MethodNotFoundException e) {
            throw new IOException(e);
        } catch (UserException e) {
            throw new IOException(e);
        } catch (SyntaxException e) {
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void twoFaces() throws IOException {

        try {
            synchronized (controlServerLock) {
                controlServer.callMethod(TWOFACES_METHOD_NAME, new XOPData());
            }
        } catch (MethodNotFoundException e) {
            throw new IOException(e);
        } catch (UserException e) {
            throw new IOException(e);
        } 
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public void startLearning(int id) throws IOException {
        try {
            synchronized (controlServerLock) {
                controlServer.callMethod(START_LEARNING_METHOD_NAME, XOPData
                        .fromString("<ID>" + id + "</ID>"));
            }
        } catch (MethodNotFoundException e) {
            throw new IOException(e);
        } catch (UserException e) {
            throw new IOException(e);
        } catch (SyntaxException e) {
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void stopLearning() throws IOException {

        try {
            synchronized (controlServerLock) {
                controlServer.callMethod(STOP_LEARNING_METHOD_NAME, XOPData
                        .fromString("<stop/>"));
            }
        } catch (MethodNotFoundException e) {
            throw new IOException(e);
        } catch (UserException e) {
            throw new IOException(e);
        } catch (SyntaxException e) {
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void reset() throws IOException {

        try {
            synchronized (controlServerLock) {
                controlServer.callMethod(RESET_METHOD_NAME, new XOPData());
            }
        } catch (MethodNotFoundException e) {
            throw new IOException(e);
        } catch (UserException e) {
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Map<String, ErrorHandlingPolicy> getRemoteServerNames() {
        Map<String, ErrorHandlingPolicy> servers =
                new HashMap<String, ErrorHandlingPolicy>();
        servers.put(activationServerName, ErrorHandlingPolicy.SELF_HEALING);
        servers.put(controlServerName, ErrorHandlingPolicy.SELF_HEALING);
        return servers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetRemoteServer(RemoteServer remoteServer) {
        if (remoteServer.getServerName().equals(activationServerName)) {
            synchronized (activationServerLock) {
                activationServer = remoteServer;
            }
        } else if (remoteServer.getServerName().equals(controlServerName)) {
            synchronized (controlServerLock) {
                controlServer = remoteServer;
            }
        } else {
            assert false : "Unknown server with name "
                    + remoteServer.getServerName() + " received.";
        }
    }

}
