package de.unibi.airobots.bonsai.xcf.sensors;

import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

import net.sf.xcf.ActiveMemory;
import net.sf.xcf.event.MemoryEvent;
import net.sf.xcf.event.MemoryEventAdapter;
import net.sf.xcf.event.MemoryEventListener;
import net.sf.xcf.memory.MemoryAction;
import net.sf.xcf.memory.MemoryException;
import net.sf.xcf.transport.Attachment;
import net.sf.xcf.transport.AttachmentNotFoundException;
import net.sf.xcf.transport.XOPData;
import net.sf.xcf.xml.XPath;
import nu.xom.Attribute;
import nu.xom.Element;
import nu.xom.ParsingException;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.core.SensorListener;
import de.unibi.airobots.bonsai.sensors.SceneGridMapSensor;
import de.unibi.airobots.bonsai.util.DataConverter;
import de.unibi.airobots.bonsai.xcf.ActiveMemoryReconnectable;
import de.unibi.airobots.btl.data.DynamicGridMap;

/**
 * A sensor to receive gridmaps from an active memory.
 * 
 * @author lziegler
 */
public class XcfSceneGridMapSensor implements ActiveMemoryReconnectable,
        SceneGridMapSensor {

    private Logger logger = Logger.getLogger(getClass());

    private String memoryName = "";
    private Object memoryLock = new Object();
    private Object sharedDataLock = new Object();
    private ActiveMemory memory = null;

    /**
     * Namespace of active memory xml documents.
     */
    public static final String NAMESPACE_DBXML =
            "http://www.sleepycat.com/2002/dbxml";

    private List<SensorListener<DynamicGridMap>> listeners =
            new LinkedList<SensorListener<DynamicGridMap>>();

    private DynamicGridMap map;

    /**
     * Own listener used to intercept the memory for person changes.
     */
    private MemoryEventListener selfListener;

    /**
     * Constructor.
     * 
     * @param memoryName
     *            name of the active memory to get the persons from
     * @param mapName
     *            name of the desired map
     */
    public XcfSceneGridMapSensor(String memoryName, String mapName) {

        this.memoryName = memoryName;

        logger.debug("Create scene memory sensor for memoryName " + memoryName);

        String xPathTag = DynamicGridMap.getBaseTag();
        String xPathAttribute = "[@uri=\"" + mapName + "\"]";
        XPath xPath = new XPath(xPathTag + xPathAttribute);
        selfListener = new MemoryEventAdapter(MemoryAction.ALL, xPath) {

            @Override
            public void handleEvent(MemoryEvent e) {
                handleMemoryEvent(e);
            }
        };
    }

    private void handleMemoryEvent(MemoryEvent e) {
        try {

            synchronized (sharedDataLock) {

                // create BTL type
                XOPData data = e.getData();
                Element root = data.getDocument().getRootElement();
                if (root == null) {
                    logger.error("Received Document has no root node");
                    return;
                }
                DynamicGridMap map = new DynamicGridMap();
                DynamicGridMap.fromElement(root, map);

                // get attachment
                try {
                    if (e.getType() != MemoryAction.REMOVE) {
                        XOPData mapData = memory.getAttachment(data);
                        Attachment mapPacket =
                                mapData.getAttachment(map.getUri());
                        byte[] bytes = mapPacket.getValue();
                        map.setGridMap(DataConverter.toFloatArray(bytes));
                    }
                } catch (MemoryException ex) {
                    logger.error("Error decoding map attachment.");
                    logger.debug(ex);
                    if (ex.getCause() != null) {
                        logger.debug(ex.getCause());
                    }
                    logger.debug("Tried to get attachment for document: "
                            + data.getDocumentAsText());
                }

                // extract xcf id
                Attribute att = root.getAttribute("id", NAMESPACE_DBXML);
                if (att == null) {
                    logger.error("Received Document has no xcf id");
                    return;
                }
                int id = Integer.parseInt(att.getValue());

                MemoryAction a = e.getType();
                if (a == MemoryAction.INSERT) {
                    logger.debug("insert id " + id);
                    this.map = map;
                } else if (e.getType() == MemoryAction.REPLACE) {
                    logger.debug("replace id " + id);
                    this.map = map;
                } else if (e.getType() == MemoryAction.REMOVE) {
                    logger.debug("remove id " + id);
                    this.map = null;
                }

                for (SensorListener<DynamicGridMap> l : listeners) {
                    l.newDataAvailable(this.map);
                }

            }

        } catch (ParsingException ex) {
            logger.error(ex.getMessage());
            logger.debug(ex);
        } catch (NumberFormatException ex) {
            logger.error("Received Document has not a valid xcf id");
            logger.debug(ex);
        } catch (AttachmentNotFoundException ex) {
            logger.error(ex.getMessage());
            logger.debug(ex);
            throw ex;
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getActiveMemoryName() {
        return memoryName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetActiveMemory(ActiveMemory activeMemory) {

        synchronized (memoryLock) {
            this.memory = activeMemory;

            try {
                memory.addListener(selfListener);
            } catch (MemoryException e) {

            }
        }
    }

    @Override
    public boolean hasNext() {
        return map != null;
    }

    @Override
    public Class<DynamicGridMap> getDataType() {
        return DynamicGridMap.class;
    }

    @Override
    public DynamicGridMap readLast() throws IOException,
            InterruptedException {
        return map;
    }

    @Override
    public DynamicGridMap readLast(long timeout) throws IOException,
            InterruptedException {
        return map;
    }

    @Override
    public void addSensorListener(SensorListener<DynamicGridMap> listener) {
        listeners.add(listener);
    }

    @Override
    public void removeSensorListener(SensorListener<DynamicGridMap> listener) {
        listeners.remove(listener);
    }

}
