package de.unibi.airobots.bonsai.xcf.sensors;

import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import net.sf.xcf.ActiveMemory;
import net.sf.xcf.event.MemoryEvent;
import net.sf.xcf.event.MemoryEventAdapter;
import net.sf.xcf.event.MemoryEventListener;
import net.sf.xcf.memory.MemoryAction;
import net.sf.xcf.memory.MemoryException;
import net.sf.xcf.xml.XPath;
import nu.xom.Attribute;
import nu.xom.Element;
import nu.xom.ParsingException;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.core.SensorListener;
import de.unibi.airobots.bonsai.sensors.SceneObjectSensor;
import de.unibi.airobots.bonsai.sensors.data.ObjectPositionList;
import de.unibi.airobots.bonsai.xcf.ActiveMemoryReconnectable;
import de.unibi.airobots.btl.data.ObjectPositionData;

/**
 * @author lziegler
 */
public class XcfSceneObjectSensor implements ActiveMemoryReconnectable,
        SceneObjectSensor {

    private Logger logger = Logger.getLogger(getClass());

    private String memoryName = "";
    private Object memoryLock = new Object();
    private Object sharedDataLock = new Object();
    private ActiveMemory memory = null;
    
    public static final String NAMESPACE_DBXML =
        "http://www.sleepycat.com/2002/dbxml";

    private List<SensorListener<ObjectPositionList>> listeners =
            new LinkedList<SensorListener<ObjectPositionList>>();

    private Map<Integer, ObjectPositionData> objects = new HashMap<Integer, ObjectPositionData>();

    /**
     * Own listener used to intercept the memory for person changes.
     */
    private MemoryEventListener selfListener;

    /**
     * Constructor.
     * 
     * @param memoryName
     *            name of the active memory to get the persons from
     */
    public XcfSceneObjectSensor(String memoryName) {

        this.memoryName = memoryName;

        logger.debug("Create scene memory sensor for memoryName " + memoryName);

        selfListener =
                new MemoryEventAdapter(MemoryAction.ALL, new XPath(
                        ObjectPositionData.getBaseTag())) {

                    @Override
                    public void handleEvent(MemoryEvent e) {
                        handleMemoryEvent(e);
                    }
                };
    }

    private void handleMemoryEvent(MemoryEvent e) {
        try {

            // create BTL type
            Element root = e.getData().getDocument().getRootElement();
            if (root == null) {
                logger.error("Received Document has no root node");
                return;
            }
            ObjectPositionData obj = new ObjectPositionData();
            ObjectPositionData.fromElement(root, obj);

            // extract xcf id
            Attribute att = root.getAttribute("id", NAMESPACE_DBXML);
            if (att == null) {
                logger.error("Received Document has no xcf id");
                return;
            }
            int id = Integer.parseInt(att.getValue());

            synchronized (sharedDataLock) {

                MemoryAction a = e.getType();
                if (a == MemoryAction.INSERT || a == MemoryAction.REPLACE) {
                    objects.put(id, obj);
                } else if (e.getType() == MemoryAction.REMOVE) {
                    objects.remove(id);
                }

                for (SensorListener<ObjectPositionList> l : listeners) {
                    l.newDataAvailable(makeList(objects.values()));
                }

            }

        } catch (ParsingException ex) {
            logger.error(ex.getMessage());
            logger.debug(ex);
        } catch (NumberFormatException ex) {
            logger.error("Received Document has not a valid xcf id");
            logger.debug(ex);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getActiveMemoryName() {
        return memoryName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetActiveMemory(ActiveMemory activeMemory) {

        synchronized (memoryLock) {
            this.memory = activeMemory;

            try {
                memory.addListener(selfListener);
            } catch (MemoryException e) {

            }
        }
    }

    @SuppressWarnings("unchecked")
    @Override
    public Class<ObjectPositionList> getDataType() {
        return (Class<ObjectPositionList>) objects.values().getClass();
    }

    @Override
    public ObjectPositionList readLast() throws IOException,
            InterruptedException {
        return makeList(objects.values());
    }

    @Override
    public ObjectPositionList readLast(long timeout) throws IOException,
            InterruptedException {
        return makeList(objects.values());
    }

    @Override
    public boolean hasNext() {
        return objects.size() > 0;
    }

    @Override
    public void addSensorListener(SensorListener<ObjectPositionList> listener) {
        listeners.add(listener);

    }

    @Override
    public void removeSensorListener(SensorListener<ObjectPositionList> listener) {
        listeners.remove(listener);
    }

    private ObjectPositionList makeList(Collection<ObjectPositionData> o) {
        ObjectPositionList list = new ObjectPositionList();
        list.addAll(o);
        return list;
    }
}
