package de.unibi.airobots.bonsai.core;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * {@link CoreObjectFactory} used for unit tests.
 * 
 * @author jwienke
 */
public class TestFactory implements CoreObjectFactory {

    public static final String FAIL_KEY = "fail";

    private Map<String, Class<?>> knownSensorKeysToType =
            new HashMap<String, Class<?>>();
    private Map<String, Class<? extends Actuator>> knownActuatorKeysToImplClass =
            new HashMap<String, Class<? extends Actuator>>();

    /**
     * {@inheritDoc}
     */
    @Override
    public void configureSensors(Set<SensorToConfigure> sensors)
            throws IllegalArgumentException, CoreObjectCreationException {

        for (SensorToConfigure s : sensors) {

            if (s.getKey() == null) {
                throw new CoreObjectCreationException("empty key");
            }
            if (s.getSensorClass() == null) {
                throw new CoreObjectCreationException("empty class");
            }

            if (knownSensorKeysToType.containsKey(s.getKey())) {
                throw new CoreObjectCreationException("duplicate key");
            }

            knownSensorKeysToType.put(s.getKey(), s.getDataTypeClass());

        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public <T> Sensor<T> createSensor(String key, Class<T> dataType)
            throws IllegalArgumentException, CoreObjectCreationException {
        if (knownSensorKeysToType.containsKey(key)
                && knownSensorKeysToType.get(key).equals(dataType)) {
            return new TestSensor<T>(dataType);
        } else {
            throw new IllegalArgumentException("Unknown key-type combination.");
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void initialize(Map<String, String> options)
            throws IllegalArgumentException {
        if (options.containsKey(FAIL_KEY)) {
            throw new IllegalArgumentException();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean canCreateSensor(String key, Class<?> dataType) {
        return knownSensorKeysToType.containsKey(key)
                && knownSensorKeysToType.get(key).equals(dataType);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void configureActuators(Set<ActuatorToConfigure> actuators)
            throws IllegalArgumentException, CoreObjectCreationException {

        for (ActuatorToConfigure a : actuators) {

            if (a.getKey() == null) {
                throw new CoreObjectCreationException("empty key");
            }
            if (a.getActuatorClass() == null) {
                throw new CoreObjectCreationException("empty class");
            }

            if (knownActuatorKeysToImplClass.containsKey(a.getKey())) {
                throw new CoreObjectCreationException("duplicate key");
            }

            knownActuatorKeysToImplClass.put(a.getKey(), a.getActuatorClass());

        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean canCreateActuator(String key,
            Class<? extends Actuator> actuatorClass) {
        return knownActuatorKeysToImplClass.containsKey(key)
                && actuatorClass.isAssignableFrom(knownActuatorKeysToImplClass
                        .get(key));
    }

    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    @Override
    public <T extends Actuator> T createActuator(String key,
            Class<T> actuatorClass) throws IllegalArgumentException,
            CoreObjectCreationException {

        if (!canCreateActuator(key, actuatorClass)) {
            throw new IllegalArgumentException(
                    "Cannot create an actuator with key '" + key + "' of type "
                            + actuatorClass);
        }

        try {
            return (T) knownActuatorKeysToImplClass.get(key).newInstance();
        } catch (InstantiationException e) {
            throw new CoreObjectCreationException(e);
        } catch (IllegalAccessException e) {
            throw new CoreObjectCreationException(e);
        }

    }

}
